/*
---------------------------------------------------------------------
-- TITLE: Ethernet DMA
-- AUTHOR: Steve Rhoads (rhoadss@yahoo.com)
-- DATE CREATED: 12/27/07
-- FILENAME: eth_dma.vhd
-- PROJECT: Plasma CPU core
-- COPYRIGHT: Software placed into the public domain by the author.
--    Software 'as is' without warranty.  Author liable for nothing.
-- DESCRIPTION:
--    Ethernet DMA (Direct Memory Access) controller.  
--    Reads four bits and writes four bits from/to the Ethernet PHY each 
--    2.5 MHz clock cycle.  Received data is DMAed starting at 0x13ff0000 
--    transmit data is read from 0x13fd0000.
--    To send a packet write bytes/4 to Ethernet send register.
---------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use work.mlite_pack.all;
*/

module eth_dma(
   clk ,
   reset,
   enable_eth,
   select_eth,
   rec_isr,
   send_isr,

   address,
   byte_we,
   data_write,
   data_read,
   pause_in,

   mem_address,
   mem_byte_we,
   data_w,
   pause_out,

   E_RX_CLK,
   E_RX_DV,
   E_RXD,
   E_TX_CLK,
   E_TX_EN,
   E_TXD);
   
input clk ;                     
input reset;
input enable_eth;                
input select_eth;
output rec_isr;   
reg rec_isr;              
output send_isr;    
reg send_isr;                

output [31:2] address;
reg [31:2] address;
output [3:0] byte_we;
reg [3:0] byte_we;
output [31:0] data_write;
reg [31:0] data_write;
input [31:0] data_read ;
input pause_in;

input [31:2] mem_address;
input [3:0] mem_byte_we;
input [31:0] data_w;
output pause_out;
reg pause_out;

input E_RX_CLK;                     
input E_RX_DV;                    
input [3:0] E_RXD ;
input E_TX_CLK;               
output E_TX_EN;     
reg E_TX_EN;             
output [3:0] E_TXD;
reg [3:0] E_TXD;


//architecture logic of eth_dma is
   reg [1:0] rec_clk;
   reg [31:0] rec_store;
   reg [27:0] rec_data ;
   reg [2:0] rec_cnt   ;
   reg [13:0] rec_words;
   reg [1:0] rec_dma   ;
   reg rec_done  ;

   reg [1:0] send_clk  ;
   reg [31:0] send_read;
   reg [31:0] send_data;
   reg [2:0] send_cnt  ;
   reg [8:0] send_words;
   reg [8:0] send_level;
   reg [1:0] send_dma  ;
   reg send_enable;
   reg temp_posedge, old_clk;

	parameter [31:0] ZERO = 32'b 00000000000000000000000000000000;   

	//)begin  --architecture

/*

   dma_proc: process(clk, reset, enable_eth, select_eth, 
         data_read, pause_in, mem_address, mem_byte_we, data_w, 
         E_RX_CLK, E_RX_DV, E_RXD, E_TX_CLK,
         rec_clk, rec_store, rec_data, 
         rec_cnt, rec_words, rec_dma, rec_done,
         send_clk, send_read, send_data, send_cnt, send_words, 
         send_level, send_dma, send_enable)
   begin

      if reset = '1' then
         rec_clk <= "00";
         rec_cnt <= "000";
         rec_words <= ZERO(13 downto 0);
         rec_dma <= "00";
         rec_done <= '0';
         send_clk <= "00";
         send_cnt <= "000";
         send_words <= ZERO(8 downto 0);
         send_level <= ZERO(8 downto 0);
         send_dma <= "00";
         send_enable <= '0';
      elsif rising_edge(clk) then

         --Receive nibble on low->high E_RX_CLK.  Send to DDR every 32 bits.
         rec_clk <= rec_clk(0) & E_RX_CLK;
         if rec_clk = "01" and enable_eth = '1' then
            if E_RX_DV = '1' or rec_cnt /= "000" then
               if rec_cnt = "111" then
                  rec_store <= rec_data & E_RXD;
                  rec_dma(0) <= '1';           --request DMA
               end if;
               rec_data <= rec_data(23 downto 0) & E_RXD;
               rec_cnt <= rec_cnt + 1;
            end if;
         end if;

         --Set transmit count or clear receive interrupt
         if select_eth = '1' then
            if mem_byte_we /= "0000" then
               send_cnt <= "000";
               send_words <= ZERO(8 downto 0);
               send_level <= data_w(8 downto 0);
               send_dma(0) <= '1';
            else
               rec_done <= '0';
            end if;
         end if;

         --Transmit nibble on low->high E_TX_CLK.  Get 32 bits from DDR.
         send_clk <= send_clk(0) & E_TX_CLK;
         if send_clk = "01" then
            if send_cnt = "111" then
               if send_words /= send_level then
                  send_data <= send_read;
                  send_dma(0) <= '1';
                  send_enable <= '1';
               else
                  send_enable <= '0';
               end if;
            else
               send_data(31 downto 4) <= send_data(27 downto 0);
            end if;
            send_cnt <= send_cnt + 1;
         end if;

         --Pick which type of DMA operation: bit0 = request; bit1 = active
         if pause_in = '0' then
            if rec_dma(1) = '1' then
               rec_dma <= "00";               --DMA done
               rec_words <= rec_words + 1;
               if E_RX_DV = '0' then
                  rec_done <= '1';
               end if;
            elsif send_dma(1) = '1' then
               send_dma <= "00";
               send_words <= send_words + 1;
               send_read <= data_read;
            elsif rec_dma(0) = '1' then
               rec_dma(1) <= '1';             --start DMA
            elsif send_dma(0) = '1' then
               send_dma(1) <= '1';            --start DMA
            end if;
         end if;

      end if;  --rising_edge(clk)

      E_TXD <= send_data(31 downto 28);
      E_TX_EN <= send_enable;
      rec_isr <= rec_done;
      if send_words = send_level then
         send_isr <= '1';
      else
         send_isr <= '0';
      end if;

      if rec_dma(1) = '1' then
         address <= "0001001111111111" & rec_words;        --0x13ff0000
         byte_we <= "1111";
         data_write <= rec_store;
         pause_out <= '1';                                 --to CPU
      elsif send_dma(1) = '1' then
         address <= "000100111111111000000" & send_words;  --0x13fe0000
         byte_we <= "0000";
         data_write <= data_w;
         pause_out <= '1';
      else
         address <= mem_address;  --Send request from CPU to DDR
         byte_we <= mem_byte_we;
         data_write <= data_w;
         pause_out <= '0';
      end if;

   end process;

*/	
	
	always @(clk)
	begin
		if(clk == 1'b0)
			temp_posedge <= 0;
		else if(clk == 1'b1 && temp_posedge == 1'b0)
			temp_posedge <= 1'b1;
		else
			temp_posedge <= 1'b0;
	
	end
	
	always @(clk, reset)//, enable_eth, select_eth,data_read, pause_in, mem_address, mem_byte_we, data_w,E_RX_CLK, E_RX_DV, E_RXD, E_TX_CLK,rec_clk, rec_store, rec_data,rec_cnt, rec_words, rec_dma, rec_done,send_clk, send_read, send_data, send_cnt, send_words,send_level, send_dma, send_enable)
	begin

		if(reset == 1'b 1)begin
			rec_clk <= 2'b 00;
			rec_cnt <= 3'b 000;
			rec_words <= ZERO[13:0];
			rec_dma <= 2'b 00;
			rec_done <= 1'b 0;
			send_clk <= 2'b 00;
			send_cnt <= 3'b 000;
			send_words <= ZERO[8:0];
			send_level <= ZERO[8:0];
			send_dma <= 2'b 00;
			send_enable <= 1'b 0;
		end
		else if ( old_clk != clk )begin
			old_clk <= clk;
			if (temp_posedge == 1'b 1 ) begin
				rec_clk <= {rec_clk[0] , E_RX_CLK};
				if(rec_clk == 2'b 01 && enable_eth == 1'b 1)begin
					if(E_RX_DV == 1'b 1 || rec_cnt != 3'b 000)begin
						if(rec_cnt == 3'b 111)begin
							rec_store <= {rec_data , E_RXD};
							rec_dma[0] <= 1'b 1;
						end
					end
					rec_data <= {rec_data[23:0] , E_RXD};
					rec_cnt <= rec_cnt + 1;
				end
			end
		end

		if(select_eth == 1'b 1 )begin
			if(mem_byte_we != 4'b 0000 )begin
				send_cnt <= 3'b 000;
				send_words <= ZERO[8:0];
				send_level <= data_w[8:0];
				send_dma[0] <= 1'b 1;
			end
			else begin
				rec_done <= 1'b 0;
			end
		end

		send_clk <= {send_clk[0] , E_TX_CLK};
		if(send_clk == 2'b 01 )begin
			if(send_cnt == 3'b 111 )begin
				if(send_words !=send_level )begin
					send_data <= send_read;
					send_dma[0] <= 1'b 1;
					send_enable <= 1'b 1;
				end
				else begin
					send_enable <= 1'b 0;
				end
			end
			else begin
				send_data[31:4] <= send_data[27:0];
			end
			send_cnt <= send_cnt + 1;
		end

		if(pause_in == 1'b 0 )begin
			if(rec_dma[1] == 1'b 1 )begin
				rec_dma <= 2'b 00;      
				rec_words <= rec_words + 1;
				if(E_RX_DV == 1'b 0 )begin
					rec_done <= 1'b 1;
				end
			end
			else if(send_dma[1] == 1'b 1 )begin
				send_dma <= 2'b 00;
				send_words <= send_words + 1;
				send_read <= data_read;
			end
			else if(rec_dma[0] == 1'b 1 )begin
				rec_dma[1] <= 1'b 1;        
			end			   
			else if(send_dma[0] == 1'b 1 )begin
				send_dma[1] <= 1'b 1;            
			end
		end

		E_TXD <= send_data[31:28];
		E_TX_EN <= send_enable;
		rec_isr <= rec_done;
		if(send_words == send_level)begin
			send_isr <= 1'b 1;
		end
		else begin
			send_isr <= 1'b 0;
		end

		if(rec_dma[1] == 1'b 1 )begin
			address <= {16'b 0001001111111111 , rec_words};       
			byte_we <= 4'b 1111;
			data_write <= rec_store;
			pause_out <= 1'b 1;                   
		end
		else if(send_dma[1] == 1'b 1 )begin
			address <= {21'b 000100111111111000000 , send_words};  
			byte_we <= 4'b 0000;
			data_write <= data_w;
			pause_out <= 1'b 1;
		end
		else begin
			address <= mem_address; 
			byte_we <= mem_byte_we;
			data_write <= data_w;
			pause_out <= 1'b 0;
		end
	end

	//end; --architecture logic
endmodule

