/*
---------------------------------------------------------------------
-- TITLE: Bus Multiplexer / Signal Router
-- AUTHOR: Steve Rhoads (rhoadss@yahoo.com)
-- DATE CREATED: 2/8/01
-- FILENAME: bus_mux.vhd
-- PROJECT: Plasma CPU core
-- COPYRIGHT: Software placed into the public domain by the author.
--    Software 'as is' without warranty.  Author liable for nothing.
-- DESCRIPTION:
--    This entity is the main signal router.  
--    It multiplexes signals from multiple sources to the correct location.
--    The outputs are as follows:
--       a_bus        : goes to the ALU
--       b_bus        : goes to the ALU
--       reg_dest_out : goes to the register bank
--       take_branch  : goes to pc_next
---------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use work.mlite_pack.all;
*/
module bus_mux(imm_in, reg_source, a_mux, a_out, reg_target, b_mux, b_out, c_bus, c_memory, c_pc, c_pc_plus4, c_mux, reg_dest_out, branch_func, take_branch);

input [15:0] imm_in;
input [31:0]    reg_source;
input   [1:0]      a_mux ;
output [31:0]    a_out;
reg [31:0]    a_out;

input [31:0]  reg_target;
input   [1:0]     b_mux ;
output [31:0] b_out ;
reg [31:0] b_out ;

input [31:0] c_bus       ;
input [31:0] c_memory     ;
input [31:2] c_pc        ;
input [31:2] c_pc_plus4   ;
input [2:0] c_mux       ;
output [31:0] reg_dest_out ;
reg [31:0] reg_dest_out ;

input [2:0]   branch_func;
output   take_branch;
reg take_branch;

/*
entity bus_mux is
   port(imm_in       : in  std_logic_vector(15 downto 0);
        reg_source   : in  std_logic_vector(31 downto 0);
        a_mux        : in  a_source_type;
        a_out        : out std_logic_vector(31 downto 0);

        reg_target   : in  std_logic_vector(31 downto 0);
        b_mux        : in  b_source_type;
        b_out        : out std_logic_vector(31 downto 0);

        c_bus        : in  std_logic_vector(31 downto 0);
        c_memory     : in  std_logic_vector(31 downto 0);
        c_pc         : in  std_logic_vector(31 downto 2);
        c_pc_plus4   : in  std_logic_vector(31 downto 2);
        c_mux        : in  c_source_type;
        reg_dest_out : out std_logic_vector(31 downto 0);

        branch_func  : in  branch_function_type;
        take_branch  : out std_logic);
end; --entity bus_mux
*/
/*
   constant A_FROM_REG_SOURCE : a_source_type := "00";
   constant A_FROM_IMM10_6    : a_source_type := "01";
   constant A_FROM_PC         : a_source_type := "10";
   
   constant C_FROM_NULL       : c_source_type := "000";
   constant C_FROM_ALU        : c_source_type := "001";
   constant C_FROM_SHIFT      : c_source_type := "001"; --same as alu
   constant C_FROM_MULT       : c_source_type := "001"; --same as alu
   constant C_FROM_MEMORY     : c_source_type := "010";
   constant C_FROM_PC         : c_source_type := "011";
   constant C_FROM_PC_PLUS4   : c_source_type := "100";
   constant C_FROM_IMM_SHIFT16: c_source_type := "101";
   constant C_FROM_REG_SOURCEN: c_source_type := "110";  

   constant BRANCH_LTZ : branch_function_type := "000";
   constant BRANCH_LEZ : branch_function_type := "001";
   constant BRANCH_EQ  : branch_function_type := "010";
   constant BRANCH_NE  : branch_function_type := "011";
   constant BRANCH_GEZ : branch_function_type := "100";
   constant BRANCH_GTZ : branch_function_type := "101";
   constant BRANCH_YES : branch_function_type := "110";
   constant BRANCH_NO  : branch_function_type := "111";
   
   
*/
   parameter [1:0] A_FROM_REG_SOURCE = 2'b 00;
   parameter [1:0] A_FROM_IMM10_6    = 2'b 01;
   parameter [1:0] A_FROM_PC         = 2'b 10;

   parameter [2:0] C_FROM_NULL       = 3'b 000;
   parameter [2:0] C_FROM_ALU        = 3'b 001;
   parameter [2:0] C_FROM_SHIFT      = 3'b 001; 
   parameter [2:0] C_FROM_MULT       = 3'b 001; 
   parameter [2:0] C_FROM_MEMORY     = 3'b 010;
   parameter [2:0] C_FROM_PC         = 3'b 011;
   parameter [2:0] C_FROM_PC_PLUS4   = 3'b 100;
   parameter [2:0] C_FROM_IMM_SHIFT16 = 3'b 101;
   parameter [2:0] C_FROM_REG_SOURCEN = 3'b 110;    

   parameter [2:0] BRANCH_LTZ = 3'b 000;
   parameter [2:0] BRANCH_LEZ = 3'b 001;
   parameter [2:0] BRANCH_EQ  = 3'b 010;
   parameter [2:0] BRANCH_NE  = 3'b 011;
   parameter [2:0] BRANCH_GEZ = 3'b 100;
   parameter [2:0] BRANCH_GTZ = 3'b 101;
   parameter [2:0] BRANCH_YES = 3'b 110;
   parameter [2:0] BRANCH_NO  = 3'b 111;  
   
   parameter [1:0] B_FROM_REG_TARGET = 2'b 00;
   parameter [1:0] B_FROM_IMM        = 2'b 01;
   parameter [1:0] B_FROM_SIGNED_IMM = 2'b 10;
   parameter [1:0] B_FROM_IMMX4      = 2'b 11;   

parameter [31:0] ZERO = 32'b 00000000000000000000000000000000;   
   
   always @(reg_source, imm_in, a_mux, c_pc)
	   case (a_mux)
	   A_FROM_REG_SOURCE :
		  a_out <= reg_source;
	   A_FROM_IMM10_6 :
		  a_out <= {ZERO[31:5] , imm_in[10:6]};
	   A_FROM_PC :
		  a_out <= {c_pc , 2'b 00};
	   default :
		  a_out <= {c_pc , 2'b 00};
	   endcase;

always @(reg_target, imm_in, b_mux) 
   case (b_mux)
    B_FROM_REG_TARGET :
      b_out <= reg_target;
    B_FROM_IMM :
      b_out <= {ZERO[31:16] , imm_in};
    B_FROM_SIGNED_IMM :
	begin
      if( imm_in[15] == 1'b 0)
	  begin
         b_out[31:16] <= ZERO[31:16];
		end
      else
	  begin
         b_out[31:16] <= 16'b 1111111111111111;
      end
      b_out[15:0] <= imm_in;
	end
    B_FROM_IMMX4 :
	begin
      if (imm_in[15] == 1'b 0)
	  begin
         b_out[31:18] <= 14'b 00000000000000;
		
	  end
      else
	  begin
         b_out[31:18] <= 14'b 11111111111111;
	  end
      b_out[17:0] <= {imm_in , 2'b 00};
	 end
   default :
      b_out <= reg_target;
   endcase;	   
	   
	   
/*
architecture logic of bus_mux is
begin
*/
/*
--Determine value of a_bus
amux: process(reg_source, imm_in, a_mux, c_pc) 
begin
   case a_mux is
   when A_FROM_REG_SOURCE =>
      a_out <= reg_source;
   when A_FROM_IMM10_6 =>
      a_out <= ZERO(31 downto 5) & imm_in(10 downto 6);
   when A_FROM_PC =>
      a_out <= c_pc & "00";
   when others =>
      a_out <= c_pc & "00";
   end case;
end process;
*/
/*
--Determine value of b_bus
bmux: process(reg_target, imm_in, b_mux) 
begin
   case b_mux is
   when B_FROM_REG_TARGET =>
      b_out <= reg_target;
   when B_FROM_IMM =>
      b_out <= ZERO(31 downto 16) & imm_in;
   when B_FROM_SIGNED_IMM =>
      if imm_in(15) = '0' then
         b_out(31 downto 16) <= ZERO(31 downto 16);
      else
         b_out(31 downto 16) <= "1111111111111111";
      end if;
      b_out(15 downto 0) <= imm_in;
   when B_FROM_IMMX4 =>
      if imm_in(15) = '0' then
         b_out(31 downto 18) <= "00000000000000";
      else
         b_out(31 downto 18) <= "11111111111111";
      end if;
      b_out(17 downto 0) <= imm_in & "00";
   when others =>
      b_out <= reg_target;
   end case;
end process;
*/

always @(c_bus, c_memory, c_pc, c_pc_plus4, imm_in, c_mux) 
   case (c_mux)
    C_FROM_ALU :
      reg_dest_out <= c_bus;
    C_FROM_MEMORY :
      reg_dest_out <= c_memory;
    C_FROM_PC :
      reg_dest_out <= {c_pc[31:2] , 2'b 00}; 
    C_FROM_PC_PLUS4 :
      reg_dest_out <= {c_pc_plus4 , 2'b 00};
    C_FROM_IMM_SHIFT16 :
      reg_dest_out <= {imm_in , ZERO[15:0]};
    default :
      reg_dest_out <= c_bus;
   endcase
/*
--Determine value of c_bus								
cmux: process(c_bus, c_memory, c_pc, c_pc_plus4, imm_in, c_mux) 
begin
   case c_mux is
   when C_FROM_ALU =>  -- | C_FROM_SHIFT | C_FROM_MULT =>
      reg_dest_out <= c_bus;
   when C_FROM_MEMORY =>
      reg_dest_out <= c_memory;
   when C_FROM_PC =>
      reg_dest_out <= c_pc(31 downto 2) & "00"; 
   when C_FROM_PC_PLUS4 =>
      reg_dest_out <= c_pc_plus4 & "00";
   when C_FROM_IMM_SHIFT16 =>
      reg_dest_out <= imm_in & ZERO(15 downto 0);
   when others =>
      reg_dest_out <= c_bus;
   end case;
end process;
*/
reg is_equal;
always @(branch_func, reg_source, reg_target) 
begin
	if( reg_source == reg_target)
		is_equal = 1'b1;
	else
		is_equal = 1'b0;
   case (branch_func)
    BRANCH_LTZ :
      take_branch <= reg_source[31];
    BRANCH_LEZ :
      take_branch <= (reg_source[31] | is_equal);
    BRANCH_EQ :
      take_branch <= is_equal;
    BRANCH_NE :
      take_branch <=  ~(is_equal);
    BRANCH_GEZ :
      take_branch <= ~(reg_source[31]);
    BRANCH_GTZ :
      take_branch <= ~reg_source[31] & ~(is_equal);
    BRANCH_YES :
      take_branch <= 1'b1;
    default :
      take_branch <= 1'b0;
   endcase
end


/*
--Determine value of take_branch
pc_mux: process(branch_func, reg_source, reg_target) 
   variable is_equal : std_logic;
begin
   if reg_source = reg_target then
      is_equal := '1';
   else
      is_equal := '0';
   end if;
   case branch_func is
   when BRANCH_LTZ =>
      take_branch <= reg_source(31);
   when BRANCH_LEZ =>
      take_branch <= reg_source(31) or is_equal;
   when BRANCH_EQ =>
      take_branch <= is_equal;
   when BRANCH_NE =>
      take_branch <= not is_equal;
   when BRANCH_GEZ =>
      take_branch <= not reg_source(31);
   when BRANCH_GTZ =>
      take_branch <= not reg_source(31) and not is_equal;
   when BRANCH_YES =>
      take_branch <= '1';
   when others =>
      take_branch <= '0';
   end case;
end process;

end; --architecture logic
*/

endmodule
