module cache(clk, reset, address_next, byte_we_next, cpu_address, mem_busy, cache_access, cache_checking, cache_miss);
input clk;
input reset;
input [31:2] address_next;
input [3:0] byte_we_next;
input [31:2] cpu_address;
input mem_busy;
output reg cache_access;
output reg cache_checking;
output reg cache_miss;

parameter STATE_IDLE     = 2'b00;
parameter STATE_CHECKING = 2'b01;
parameter STATE_MISSED   = 2'b10;
parameter STATE_WAITING  = 2'b11;
parameter ZERO          		= 32'b00000000000000000000000000000000;
parameter ADDRESS_WIDTH = 10;

reg [8:0] storage [0:2 ** ADDRESS_WIDTH - 1];

reg [1:0] state_reg;
reg [1:0] state;
reg [1:0] state_next;
reg [10:0] cache_address;
reg [8:0] cache_tag_in;
reg [8:0] cache_tag_reg;
reg [8:0] cache_tag_out;
reg cache_we;

always @(posedge clk)
begin
		if (reset == 1'b1)
		begin
			state_reg = STATE_IDLE;
			cache_tag_reg = ZERO[8:0];
		end
		else
		begin
			state_reg = state_next;
			if (state == STATE_IDLE && state_reg != STATE_MISSED)
				cache_tag_reg = cache_tag_in;
		end
end
	always @(state_reg, cache_tag_out, cache_tag_reg, mem_busy) 
	begin
		case (state_reg)
		STATE_IDLE :      //      --cache idle
		begin
			cache_checking = 1'b0;
			cache_miss = 1'b0; 
			state = STATE_IDLE;
		end
		STATE_CHECKING:      //  --current read in cached range, check if match
		begin
			cache_checking = 1'b1;
			if ((cache_tag_out != cache_tag_reg) || (cache_tag_out == 9'b111111111))
			begin
				cache_miss = 1'b1;
				state = STATE_MISSED;
			end
			else
			begin
				cache_miss = 1'b0;
				state = STATE_IDLE;
			end
		end
		STATE_MISSED:        // --current read cache miss
		begin
			cache_checking = 1'b0;
			cache_miss = 1'b1;
			if (mem_busy == 1'b1)
				state = STATE_MISSED;
			else
				state = STATE_WAITING;
		end	
		STATE_WAITING:    //     --waiting for memory access to complete
		begin
			cache_checking = 1'b0;
			cache_miss = 1'b0;
			if (mem_busy == 1'b1)
				state = STATE_WAITING;
			else
				state = STATE_IDLE;
		end
		default:
		begin
			cache_checking = 1'b0;
			cache_miss = 1'b0;
			state = STATE_IDLE;
		end
		endcase
	end
	always @(state, address_next, byte_we_next, cpu_address)
	begin
		if (state == STATE_IDLE)  // --check if next access in cached range
		begin
			cache_address = {1'b0 , address_next[11:2]};
			if (address_next[30:21] == 10'b0010000000) // --first 2MB of DDR
            begin
				cache_access = 1'b1;
				if (byte_we_next == 4'b0000)    // --read cycle
				begin
					cache_we = 1'b0;
					state_next = STATE_CHECKING;  //--need to check if match
				end
				else
				begin
					cache_we = 1'b1;    //           --update cache tag
					state_next = STATE_WAITING;
				end
			end
			else
			begin
				cache_access = 1'b0;
				cache_we = 1'b0;
				state_next = STATE_IDLE;
			end
		end
		else
		begin
			cache_address = {1'b0, cpu_address[11:2]};
			cache_access = 1'b0;
			if (state == STATE_MISSED)
				cache_we = 1'b1;         //         --update cache tag
			else
				cache_we = 1'b0;
			state_next = state;
		end

		if (byte_we_next == 4'b0000 || byte_we_next == 4'b1111) // --read or 32-bit write
			cache_tag_in = address_next[20:12];
		else
			cache_tag_in = 9'b111111111; // --invalid tag
	end	

    always @(posedge clk)
	begin
		if (cache_we == 1'b1)
            storage[cache_address[ADDRESS_WIDTH-1:0]] = cache_tag_in;
        cache_tag_out = storage[cache_address[ADDRESS_WIDTH-1:0]];
	end
      
endmodule

