/*
////////////////////////////////////////////////////////////////////-
// TITLE: Pipeline
// AUTHOR: Steve Rhoads (rhoadss@yahoo.com)
// DATE CREATED: 6/24/02
// FILENAME: pipeline.vhd
// PROJECT: Plasma CPU core
// COPYRIGHT: Software placed into the public domain by the author.
//    Software 'as is' without warranty.  Author liable for nothing.
// DESCRIPTION:
//    Controls the three stage pipeline by delaying the wires:
//      a_bus, b_bus, alu/shift/mult_func, c_source, and rs_index.
////////////////////////////////////////////////////////////////////-
library ieee;
use ieee.std_logic_1164.all;
use work.mlite_pack.all;

//Note: sigD <= sig after rising_edge(clk)
*/
module pipeline(clk            ,
                reset          ,
                a_bus          ,
                a_busD         ,
                b_bus          ,
                b_busD         ,
                alu_func       ,
                alu_funcD      ,
                shift_func     ,
                shift_funcD    ,
                mult_func      ,
                mult_funcD     ,
                reg_dest       ,
                reg_destD      ,
                rd_index       ,
                rd_indexD      ,
                rs_index       ,
                rt_index       ,
                pc_source      ,
                mem_source     ,
                a_source       ,
                b_source       ,
                c_source       ,
                c_bus          ,
                pause_any      ,
                pause_pipeline );

input			clk            ;//in  std_logic;
input			reset          ;//in  std_logic;
input [31:0]			a_bus          ;//in  std_logic_vector(31 downto 0);
output reg [31:0]			a_busD         ;//out std_logic_vector(31 downto 0);
input 	[31:0]		b_bus          ;//in  std_logic_vector(31 downto 0);
output reg	[31:0]		b_busD         ;//out std_logic_vector(31 downto 0);
input	[3:0]		alu_func       ;//in  alu_function_type;
output reg	[3:0]		alu_funcD      ;//out alu_function_type;
input	[1:0]		shift_func     ;//in  shift_function_type;
output reg	[1:0]		shift_funcD    ;//out shift_function_type;
input	[3:0]		mult_func      ;//in  mult_function_type;				
output reg	[3:0]		mult_funcD     ;//out mult_function_type;				
input	[31:0]		reg_dest       ;//in  std_logic_vector(31 downto 0);				
output 	[31:0]		reg_destD      ;//out std_logic_vector(31 downto 0);				
input	[5:0]		rd_index       ;//in  std_logic_vector(5 downto 0);				
output 	[5:0]		rd_indexD      ;//out std_logic_vector(5 downto 0);				
							
input	[5:0]		rs_index       ;//in  std_logic_vector(5 downto 0);				
input	[5:0]		rt_index       ;//in  std_logic_vector(5 downto 0);				
input			pc_source      ;//in  pc_source_type;				
input			mem_source     ;//in  mem_source_type;				
input	[1:0]		a_source       ;//in  a_source_type;				
input	[1:0]		b_source       ;//in  b_source_type;				
input	[2:0]		c_source       ;//in  c_source_type;				
input	[31:0]		c_bus          ;//in  std_logic_vector(31 downto 0);				
input			pause_any      ;//in  std_logic;				
output			pause_pipeline ;//out std_logic);			



/*
entity pipeline is
   port(clk            : in  std_logic;
        reset          : in  std_logic;
        a_bus          : in  std_logic_vector(31 downto 0);
        a_busD         : out std_logic_vector(31 downto 0);
        b_bus          : in  std_logic_vector(31 downto 0);
        b_busD         : out std_logic_vector(31 downto 0);
        alu_func       : in  alu_function_type;
        alu_funcD      : out alu_function_type;
        shift_func     : in  shift_function_type;
        shift_funcD    : out shift_function_type;
        mult_func      : in  mult_function_type;
        mult_funcD     : out mult_function_type;
        reg_dest       : in  std_logic_vector(31 downto 0);
        reg_destD      : out std_logic_vector(31 downto 0);
        rd_index       : in  std_logic_vector(5 downto 0);
        rd_indexD      : out std_logic_vector(5 downto 0);

        rs_index       : in  std_logic_vector(5 downto 0);
        rt_index       : in  std_logic_vector(5 downto 0);
        pc_source      : in  pc_source_type;
        mem_source     : in  mem_source_type;
        a_source       : in  a_source_type;
        b_source       : in  b_source_type;
        c_source       : in  c_source_type;
        c_bus          : in  std_logic_vector(31 downto 0);
        pause_any      : in  std_logic;
        pause_pipeline : out std_logic);
end; //entity pipeline
*/

   //-----------------------------------------
   
   parameter  ZERO          = 32'b00000000000000000000000000000000;
   parameter  ONES          = 32'b11111111111111111111111111111111;
   
   parameter  HIGH_Z        = 32'bZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ;

   parameter  ALU_NOTHING   = 4'b0000;
   parameter  ALU_ADD       = 4'b0001;
   parameter  ALU_SUBTRACT  = 4'b0010;
   parameter  ALU_LESS_THAN = 4'b0011;
   parameter  ALU_LESS_THAN_SIGNED = 4'b0100;
   parameter  ALU_OR        = 4'b0101;
   parameter  ALU_AND       = 4'b0110;
   parameter  ALU_XOR       = 4'b0111;
   parameter  ALU_NOR       = 4'b1000;

   parameter  SHIFT_NOTHING        = 2'b00;
   parameter  SHIFT_LEFT_UNSIGNED  = 2'b01;
   parameter  SHIFT_RIGHT_SIGNED   = 2'b11;
   parameter  SHIFT_RIGHT_UNSIGNED = 2'b10;

   parameter  MULT_NOTHING       = 4'b0000;
   parameter  MULT_READ_LO       = 4'b0001;
   parameter  MULT_READ_HI       = 4'b0010;
   parameter  MULT_WRITE_LO      = 4'b0011;
   parameter  MULT_WRITE_HI      = 4'b0100;
   parameter  MULT_MULT          = 4'b0101;
   parameter  MULT_SIGNED_MULT   = 4'b0110;
   parameter  MULT_DIVIDE        = 4'b0111;
   parameter  MULT_SIGNED_DIVIDE = 4'b1000;

   parameter  A_FROM_REG_SOURCE = 2'b00;
   parameter  A_FROM_IMM10_6    = 2'b01;
   parameter  A_FROM_PC         = 2'b10;

   parameter  B_FROM_REG_TARGET = 2'b00;
   parameter  B_FROM_IMM        = 2'b01;
   parameter  B_FROM_SIGNED_IMM = 2'b10;
   parameter  B_FROM_IMMX4      = 2'b11;

   parameter  C_FROM_NULL       = 3'b000;
   parameter  C_FROM_ALU        = 3'b001;
   parameter  C_FROM_SHIFT      = 3'b001;
   parameter  C_FROM_MULT       = 3'b001; 
   parameter  C_FROM_MEMORY     = 3'b010;
   parameter  C_FROM_PC         = 3'b011;
   parameter  C_FROM_PC_PLUS4   = 3'b100;
   parameter  C_FROM_IMM_SHIFT16= 3'b101;
   parameter  C_FROM_REG_SOURCEN= 3'b110;


   parameter  FROM_INC4       = 2'b00;
   parameter  FROM_OPCODE25_0 = 2'b01;
   parameter  FROM_BRANCH     = 2'b10;
   parameter  FROM_LBRANCH    = 2'b11;


   parameter  BRANCH_LTZ = 3'b000;
   parameter  BRANCH_LEZ = 3'b001;
   parameter  BRANCH_EQ  = 3'b010;
   parameter  BRANCH_NE  = 3'b011;
   parameter  BRANCH_GEZ = 3'b100;
   parameter  BRANCH_GTZ = 3'b101;
   parameter  BRANCH_YES = 3'b110;
   parameter  BRANCH_NO  = 3'b111;


   parameter  MEM_FETCH   = 4'b0000;
   parameter  MEM_READ32  = 4'b0100;
   parameter  MEM_WRITE32 = 4'b0101;
   parameter  MEM_READ16  = 4'b1000;
   parameter  MEM_READ16S = 4'b1010;
   parameter  MEM_WRITE16 = 4'b1001;
   parameter  MEM_READ8   = 4'b1100;
   parameter  MEM_READ8S  = 4'b1110;
   parameter  MEM_WRITE8  = 4'b1101;  
   

   //----------------------------------------
//architecture logic of pipeline is
   reg [5:0] rd_index_reg    ;//std_logic_vector(5 downto 0);
   reg [31:0] reg_dest_reg    ;//std_logic_vector(31 downto 0);
   wire [31:0] reg_dest_delay  ;//std_logic_vector(31 downto 0);
   reg c_source_reg    ;//c_source_type;
   reg pause_enable_reg;//std_logic;
   wire pause_mult_clock;
   wire freeze_pipeline;


assign pause_mult_clock = (((pc_source != FROM_INC4) && (pc_source != FROM_OPCODE25_0)) || (mem_source != MEM_FETCH) || (mult_func == MULT_READ_LO || mult_func == MULT_READ_HI))? 1'b1:1'b0;
assign reg_dest_delay = (c_source_reg == C_FROM_ALU)? c_bus:reg_dest_reg;
assign   freeze_pipeline = ~(pause_mult_clock && pause_enable_reg) && pause_any;
assign   pause_pipeline = pause_mult_clock && pause_enable_reg;
assign   rd_indexD = rd_index_reg;
assign    reg_destD = reg_dest_delay;
   
always @(posedge(clk))
begin

	if (reset == 1'b1) 
	begin
		a_busD <= ZERO;
		b_busD <= ZERO;
		alu_funcD <= ALU_NOTHING;
		shift_funcD <= SHIFT_NOTHING;
		mult_funcD <= MULT_NOTHING;
		reg_dest_reg <= ZERO;
		c_source_reg <= "000";
		rd_index_reg <= "000000";
		pause_enable_reg <= 1'b0;
	end
	else  
	begin
		if (freeze_pipeline == 1'b0) 
		begin
			if ((rs_index == "000000" || rs_index != rd_index_reg) || (a_source != A_FROM_REG_SOURCE || pause_enable_reg == 1'b0)) 
			begin
			a_busD <= a_bus;
			end
			else
			begin
				a_busD <= reg_dest_delay;  //rs from previous operation (bypass stage)
			end

			if ((rt_index == "000000" || rt_index != rd_index_reg) || (b_source != B_FROM_REG_TARGET || pause_enable_reg == 1'b0)) begin
				b_busD <= b_bus;
			end
			else
			begin
				b_busD <= reg_dest_delay;  //rt from previous operation
			end

			alu_funcD <= alu_func;
			shift_funcD <= shift_func;
			mult_funcD <= mult_func;
			reg_dest_reg <= reg_dest;
			c_source_reg <= c_source;
			rd_index_reg <= rd_index;
		end

		if ((pause_enable_reg == 1'b0) && (pause_any == 1'b0)) 
		begin
			pause_enable_reg <= 1'b1;   //enable pause_pipeline
		end
		else if (pause_mult_clock == 1'b1) 
		begin
			pause_enable_reg <= 1'b0;   //disable pause_pipeline
		end
		
	end

end

endmodule //logic
